package com.sillycycle.bagleyd.mlink;

/*
 * @(#)MlinkSolve.java
 *
 * Taken from the algorithms in Ideal's you are about to find the
 * solution: The Missing Link (also looked at James G. Nourse's
 * The Simple Solutions to Cubic Puzzles).
 * Many speed-ups possible, for example: add more special cases,
 * use 2 ignored, speed up for non-oriented and non-middle, handle
 * reverse better so it does not repeat last move of 3 mlink.tiles.
 *
 * Copyright 2003 - 2019  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>MlinkSolve</code> class is the code that does the work
 * of solving the puzzle.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

class MlinkSolve {
	private static JFrame frame;
	private static MlinkCanvas mlink;
	//private static int counter; // debugging
	private static final boolean DEBUG = false;
	private static Random generator;

	MlinkSolve(JFrame frame, MlinkCanvas mlink) {
		MlinkSolve.frame = frame;
		MlinkSolve.mlink = mlink;
		generator = new Random(System.nanoTime());
		//this.counter = 0;
	}

	boolean moveATile(int direction, int tile, boolean shift) {
		if (!getFlag()) {
			return false;
		}
		/*if (direction % 2 == 1)
			System.out.println("move: " + direction + ", " + shift);
		else
			System.out.println("move: " + direction + ", " + tile + ", " + shift);*/
		return mlink.movePuzzle(direction, tile, shift, false,
			MlinkCanvas.DOUBLE);
	}

	static int nextInt(int n) {
		if (DEBUG) /* make it deterministic */
			return 1;
		return generator.nextInt(n);
	}

	void leftAway(boolean reverse) {
		/* push top away */
		moveATile(MlinkCanvas.TOP, (reverse) ?
			mlink.tiles - 1 : 0, false);
	}

	void leftAway2(boolean reverse) {
		/* push top away */
		moveATile(MlinkCanvas.TOP, (reverse) ?
			mlink.tiles - 1 : 0, false);
		moveATile(MlinkCanvas.TOP, (reverse) ?
			mlink.tiles - 1 : 0, false);
	}

	void leftTowards(boolean reverse) {
		/* push top towards */
		moveATile(MlinkCanvas.BOTTOM, (reverse) ?
			mlink.tiles - 1 : 0, false);
	}

	void leftTowards2(boolean reverse) {
		/* push top towards */
		moveATile(MlinkCanvas.BOTTOM, (reverse) ?
			mlink.tiles - 1 : 0, false);
		moveATile(MlinkCanvas.BOTTOM, (reverse) ?
			mlink.tiles - 1 : 0, false);
	}

	void rightAway(boolean reverse) {
		/* push top away */
		moveATile(MlinkCanvas.TOP, (reverse) ?
			0 : mlink.tiles - 1, false);
	}

	void rightAway2(boolean reverse) {
		/* push top away */
		moveATile(MlinkCanvas.TOP, (reverse) ?
			0 : mlink.tiles - 1, false);
		moveATile(MlinkCanvas.TOP, (reverse) ?
			0 : mlink.tiles - 1, false);
	}

	void rightTowards(boolean reverse) {
		/* push top towards */
		moveATile(MlinkCanvas.BOTTOM, (reverse) ?
			0 : mlink.tiles - 1, false);
	}

	void rightTowards2(boolean reverse) {
		/* push top towards */
		moveATile(MlinkCanvas.BOTTOM, (reverse) ?
			0 : mlink.tiles - 1, false);
		moveATile(MlinkCanvas.BOTTOM, (reverse) ?
			0 : mlink.tiles - 1, false);
	}

	void leftHalf(boolean reverse) {
		if (nextInt(2) == 1) {
			leftTowards2(reverse);
		} else {
			leftAway2(reverse);
		}
	}

	void rightHalf(boolean reverse) {
		if (nextInt(2) == 1) {
			rightTowards2(reverse);
		} else {
			rightAway2(reverse);
		}
	}

	void leftSlide(int n, boolean reverse) {
		if (reverse)
			moveATile(MlinkCanvas.RIGHT,
				mlink.spacePosition - n, false);
		else
			moveATile(MlinkCanvas.LEFT,
				mlink.spacePosition + n, false);
	}

	void rightSlide(int n, boolean reverse) {
		if (reverse)
			moveATile(MlinkCanvas.LEFT,
				mlink.spacePosition + n, false);
		else
			moveATile(MlinkCanvas.RIGHT,
				mlink.spacePosition - n, false);
	}


	/* Page 6 */
	void swap02_32(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);

		rightTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 2, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap01_31(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightHalf(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);

		rightTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);

		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Page 7 */
	void swap00_30(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);

		rightAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);

		rightAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	void swap02_22(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		rightHalf(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);

		rightHalf(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);

		rightHalf(reverse);
		leftSlide(2, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	/* Page 8 */
	void swap01_21(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightHalf(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);

		rightTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);

		rightHalf(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);

		rightAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse); /* Extra non-documented move */
		}
	}

	void swap00_21(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		rightHalf(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);

		rightHalf(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);

		rightHalf(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	/* Page 9 */
	/* From here on "All" is simplified to a single tile move */
	void swap32_30(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	void swap32_21(boolean reverse) {

		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap32_31(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Page 10 */
	void swap30_31(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap31_20(boolean reverse) {

		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	void swap21_20(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Page 11 */
	void swap31_22(boolean reverse) {

		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse); /* not Towards */
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	void swap30_21(boolean reverse) {

		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Page 12 */
	void swap32_20(boolean reverse) {

		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap30_22(boolean reverse) {

		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Page 13 */
	void swap22_32(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap21_31(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap20_30(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Page 14 */
	void swap22_21(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	/* Slight efficiency gain if used */
	/*
	void swap32_31_30(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	void swap32_30_31(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}
	*/

	/* move* and alt* From Simple Solutions to Cubic Puzzles */
	/*
	// 1A
	void moveRight2ToLeft2() {
		const boolean reverse = false;

		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(1, reverse);
		leftAway(reverse);
		rightSlide(1, reverse);
		leftTowards(reverse);
		leftSlide(1, reverse);
	}

	// 1B
	void moveRightMiddleToLeft() {
		const boolean reverse = false;

		rightTowards(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(1, reverse);
		leftAway(reverse);
		rightSlide(1, reverse);
		leftTowards(reverse);
		leftSlide(1, reverse);
	}

	// 1C
	void moveLeftMiddleToLeft() {
		const boolean reverse = false;

		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(1, reverse);
		leftAway(reverse);
		rightSlide(1, reverse);
		leftAway(reverse);
		leftSlide(1, reverse);
	}

	// 2E (2B should be opposite but lists the same moves)
	void moveRightMiddleToLeftMiddle() {
		const boolean reverse = false;

		rightSlide(mlink.tiles - 1, reverse);
		leftAway(reverse);
		leftSlide(mlink.tiles - 1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(mlink.tiles - 1, reverse);
		leftTowards(reverse);
		leftSlide(mlink.tiles - 1, reverse);
	}

	// altSwap3?_3? same number of moves as Swap3?_3?
	// 4A
	void altSwap30_31(boolean reverse) {
		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	// 4B
	void altSwap32_30(boolean reverse) {
		if (reverse) {
			rightSlide(mlink.tiles - 1, !reverse);
		}
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightHalf(reverse);

		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	// 4C
	void altSwap32_31(boolean reverse) {
		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightAway(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		if (reverse) {
			leftSlide(mlink.tiles - 1, !reverse);
		}
	}

	// altSwap3?_3?_3? 4 moves more than Swap3?_3?_3?
	// (2 single moves + 2 from an extra rightHalf counts as 2)
	// 4D1
	void altSwap32_30_31(boolean reverse) {
		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightAway(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}

	// 4D2
	void altSwap32_31_30(boolean reverse) {
		if (reverse) {
			rightSlide(2, !reverse);
		} else {
			rightSlide(1, reverse);
		}
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		leftSlide(1, reverse);
		rightAway(reverse);
		rightSlide(1, reverse);
		rightTowards(reverse);
		leftSlide(1, reverse);
		rightTowards(reverse);
		rightSlide(1, reverse);
		rightHalf(reverse);
		if (reverse) {
			leftSlide(2, !reverse);
		} else {
			leftSlide(1, reverse);
		}
	}
*/

	static int findPiece(int match) {
		int pos = -1;

		for (int tile = 0; tile < mlink.tiles; tile++) {
			for (int face = 0; face < mlink.faces; face++) {
				if (mlink.tileOfPosition[face * mlink.tiles + tile] - 1 == match) {
					pos = face * mlink.tiles + tile;
					break;
				}
			}
			if (pos != -1)
				break;
		}
		return pos;
	}

	/* Page 4 - 5, but did not really use it too much */
	void solveWhite() {
		boolean reverse = false;
		int current, loc, pos;

		current = (mlink.faces - 1) * mlink.tiles;
		if (mlink.tileOfPosition[current] - 1 != current) {
			loc = findPiece(current);
			if (loc % mlink.tiles != 0) { // not on left side
				if (mlink.spacePosition / mlink.tiles !=
						loc / mlink.tiles) { // not on same face
					if (mlink.spacePosition % mlink.tiles > 0) {
						rightSlide(mlink.spacePosition %
							mlink.tiles, reverse);
					}
					pos = ((loc / mlink.tiles -
						mlink.spacePosition / mlink.tiles) +
						mlink.faces) % mlink.faces;
System.out.println("before" + pos);
					switch (pos) {
					case 1:
						leftTowards(reverse);
						break;
					case 2:
						leftHalf(reverse);
						break;
					case 3:
						leftAway(reverse);
						break;
					default:
						break;
					}
				}
				if (mlink.spacePosition % mlink.tiles < mlink.tiles - 1) {
					if (loc > mlink.spacePosition)
						loc--;
					leftSlide((mlink.spacePosition / mlink.tiles) * mlink.tiles
						+ mlink.tiles - 1 - mlink.spacePosition, reverse);
				}
			}
			while (loc % mlink.tiles != 0) {
				rightAway(reverse);
				rightSlide(mlink.tiles - 1, reverse);
				leftTowards(reverse);
				leftSlide(mlink.tiles - 1, reverse);
				loc--;
			}
			pos = (loc / mlink.tiles) % mlink.faces;
System.out.println("this:" + pos);
			switch (pos) {
			case 0:
				leftAway(reverse);
				break;
			case 1:
				leftHalf(reverse);
				break;
			case 2:
				leftTowards(reverse);
				break;
			default:
				break;
			}
		}
		if (mlink.spacePosition % mlink.tiles < mlink.tiles - 1) {
			leftSlide(((mlink.spacePosition / mlink.tiles) *
				mlink.tiles + mlink.tiles - 1) -
				mlink.spacePosition, reverse);
		}
		pos = (mlink.spacePosition / mlink.tiles) % mlink.faces;
		switch (pos) {
		case 0:
			rightAway(reverse);
			break;
		case 1:
			rightHalf(reverse);
			break;
		case 2:
			rightTowards(reverse);
			break;
		default:
			break;
		}

		current++;
		if (mlink.tileOfPosition[current] - 1 != current) {
			loc = findPiece(current);
			if (mlink.faces - 1 !=
					loc / mlink.tiles) { // not same face
				if ((loc + 1) % mlink.tiles != 0) { // not right side
					pos = loc / mlink.tiles;
					switch (pos) {
					case 0:
						rightTowards(reverse);
						break;
					case 1:
						rightHalf(reverse);
						break;
					case 2:
						rightAway(reverse);
						break;
					default:
						break;
					}
					for (int i = 0; i < mlink.tiles - 1 - (loc % mlink.tiles); i++) {
						rightSlide(mlink.tiles - 1, reverse);
						leftAway(reverse);
						leftSlide(mlink.tiles - 1, reverse);
						rightTowards(reverse);
						leftTowards(reverse);
					}
					loc = (mlink.spacePosition + mlink.tiles) % mlink.tileFaces;
				}
				pos = mlink.spacePosition / mlink.tiles;
				switch (pos) {
				case 0:
					rightAway(reverse);
					loc = (loc + 3 * mlink.tiles) % mlink.tileFaces;
					break;
				case 1:
					rightHalf(reverse);
					loc = (loc + 2 * mlink.tiles) % mlink.tileFaces;
					break;
				case 2:
					rightTowards(reverse);
					loc = (loc + mlink.tiles) % mlink.tileFaces;
					break;
				default:
					break;
				}
				rightSlide(1, reverse);
				pos = loc / mlink.tiles;
				switch (pos) {
				case 0:
					rightAway(reverse);
					break;
				case 1:
					rightHalf(reverse);
					break;
				case 2:
					rightTowards(reverse);
					break;
				default:
					break;
				}
				leftSlide(1, reverse);
			}
			rightAway(reverse);
			rightSlide(mlink.tiles - 1, reverse);
			leftTowards(reverse);
			leftSlide(mlink.tiles - 1, reverse);
			leftAway(reverse);
		}

		current++;
		if (mlink.tileOfPosition[current] - 1 != current) {
			loc = findPiece(current);
			if ((loc + 1) % mlink.tiles != 0) { // not right side
				pos = loc / mlink.tiles;
				switch (pos) {
				case 0:
					rightTowards(reverse);
					break;
				case 1:
					rightHalf(reverse);
					break;
				case 2:
					rightAway(reverse);
					break;
				default:
					break;
				}
				if (pos == 0) {
				for (int i = 0; i < mlink.tiles - 1 - (loc % mlink.tiles); i++) {
					rightSlide(mlink.tiles - 1, reverse);
					leftTowards(reverse);
					leftSlide(mlink.tiles - 1, reverse);
					rightAway(reverse);
					leftAway(reverse);
				}
				loc = (mlink.spacePosition + 3 * mlink.tiles) % mlink.tileFaces;
				} else {
				for (int i = 0; i < mlink.tiles - 1 - (loc % mlink.tiles); i++) {
					rightSlide(mlink.tiles - 1, reverse);
					leftAway(reverse);
					leftSlide(mlink.tiles - 1, reverse);
					rightTowards(reverse);
					leftTowards(reverse);
				}
				loc = (mlink.spacePosition + mlink.tiles) % mlink.tileFaces;
				}
			}
			pos = mlink.spacePosition / mlink.tiles;
			switch (pos) {
			case 0:
				rightAway(reverse);
				loc = (loc + 3 * mlink.tiles) % mlink.tileFaces;
				break;
			case 1:
				rightHalf(reverse);
				loc = (loc + 2 * mlink.tiles) % mlink.tileFaces;
				break;
			case 2:
				rightTowards(reverse);
				loc = (loc + mlink.tiles) % mlink.tileFaces;
				break;
			default:
				break;
			}
			rightSlide(1, reverse);
			pos = loc / mlink.tiles;
			switch (pos) {
			case 0:
				rightAway(reverse);
				break;
			case 1:
				rightHalf(reverse);
				break;
			case 2:
				rightTowards(reverse);
				break;
			default:
				break;
			}
			leftSlide(1, reverse);
		}
	}

	/* Page 6 - 14 */
	boolean solveSpecialCases(int pos1, int pos2) {
		boolean reverse = false;
		int x_1 = pos1 % mlink.tiles;
		int y_1 = pos1 / mlink.tiles;
		int x_2 = pos2 % mlink.tiles;
		int y_2 = pos2 / mlink.tiles;

		if (DEBUG) {
			System.out.println("solveSpecialCases pos "
				+ pos1 + ", " + pos2);
		}
		if (x_1 >= mlink.tiles / 2) {
			x_1 = mlink.tiles - 1 - x_1;
			x_2 = mlink.tiles - 1 - x_2;
			reverse = !reverse;
		}
		if (x_1 > x_2 || (x_1 == x_2 && y_1 > y_2)) {
			int temp;

			temp = x_1;
			x_1 = x_2;
			x_2 = temp;
			temp = y_1;
			y_1 = y_2;
			y_2 = temp;
		}
		switch (x_1) {
		case 0:
			switch (y_1) {
			case 0:
				switch (x_2) {
				case 0:
					switch (y_2) {
					case 1:
						swap30_31(!reverse); /* 00_01 */
						break;
					case 2:
						swap32_30(!reverse); /* 00_02 */
						break;
					default:
						return false;
					}
					break;
				case 1:
					switch (y_2) {
					case 0:
						swap20_30(!reverse); /* 00_10 */
						break;
					case 1:
						swap30_21(!reverse); /* 00_11 */
						break;
					case 2:
						swap30_22(!reverse); /* 00_12 */
						break;
					default:
						return false;
					}
					break;
				case 2:
					if (y_2 == 1) {
						swap00_21(reverse);
					} else {
						return false;
					}
					break;
				case 3:
					if (y_2 == 0) {
						swap00_30(reverse);
					} else {
						return false;
					}
					break;
				default:
					return false;
				}
				break;
			case 1:
				switch (x_2) {
				case 0:
					if (y_2 == 2) {
						swap32_31(!reverse); /* 01_02 */
					} else {
						return false;
					}
					break;
				case 1:
					switch (y_2) {
					case 0:
						swap31_20(!reverse); /* 01_10 */
						break;
					case 1:
						swap21_31(!reverse); /* 01_11 */
						break;
					case 2:
						swap31_22(!reverse); /* 01_12 */
						break;
					default:
						return false;
					}
					break;
				case 2:
					if (y_2 == 1) {
						swap01_21(reverse);
					} else {
						return false;
					}
					break;
				case 3:
					if (y_2 == 1) {
						swap01_31(reverse);
					} else {
						return false;
					}
					break;
				default:
					return false;
				}
				break;
			case 2:
				switch (x_2) {
				case 1:
					switch (y_2) {
					case 0:
						swap32_20(!reverse); /* 02_10 */
						break;
					case 1:
						swap32_21(!reverse); /* 02_11 */
						break;
					case 2:
						swap22_32(!reverse); /* 02_12 */
						break;
					default:
						return false;
					}
					break;
				case 2:
					if (y_2 == 2) {
						swap02_22(reverse);
					} else {
						return false;
					}
					break;
				case 3:
					if (y_2 == 2) {
						swap02_32(reverse);
					} else {
						return false;
					}
					break;
				default:
					return false;
				}
				break;
			default:
				return false;
			}
			break;
		case 1:
			switch (y_1) {
			case 0:
				if (x_2 == 1) {
					if (y_2 == 1) {
						swap21_20(!reverse); /* 10_11 */
					} else {
						return false;
					}
				} else {
					return false;
				}
				break;
			case 1:
				if (x_2 == 1) {
					if (y_2 == 2) {
						swap22_21(!reverse); /* 11_12 */
					} else {
						return false;
					}
				} else {
					return false;
				}
				break;
			default:
				return false;
			}
			break;
		default:
			return false;
		}
		return true;
	}

	void solveMiddleFromEnd(int tile, int otherTile) {
		boolean found = true;

		while (found) {
			found = false;
			for (int face = 0; face < mlink.faces - 1; face++) {
				int pos = face * mlink.tiles + tile;
				for (int otherFace = 0; otherFace < mlink.faces - 1;
						otherFace++) {
					int otherPos = otherFace * mlink.tiles + otherTile;
					if (mlink.tileOfPosition[pos] - 1 == otherPos) {
						found = solveSpecialCases(pos, otherPos);
					}
				}
			}
		}
	}

	void solveEnd(int tile) {
		boolean found = true;

		while (found) {
			found = false;
			for (int face = 0; face < mlink.faces - 1; face++) {
				int pos = face * mlink.tiles + tile;
				for (int otherFace = 0; otherFace < mlink.faces - 1;
						otherFace++) {
					int otherPos = otherFace * mlink.tiles + tile;
					if (pos != otherPos &&
							mlink.tileOfPosition[pos] - 1 == otherPos) {
						found = solveSpecialCases(pos, otherPos);
					}
				}
			}
		}
	}

	void solveEnds(int tile, int otherTile) {
		boolean found = true;

		while (found) {
			found = false;
			for (int face = 0; face < mlink.faces - 1; face++) {
				int pos = face * mlink.tiles + tile;
				int otherPos = face * mlink.tiles + otherTile;
				if (mlink.tileOfPosition[pos] - 1 == otherPos) {
					found = solveSpecialCases(pos, otherPos);
				}
			}
		}
	}

	void matchLeftRight(int tile, int otherTile) {
		boolean found = true;

		while (found) {
			found = false;
			for (int face = 0; face < mlink.faces - 1; face++) {
				int pos = face * mlink.tiles + tile;
				for (int otherFace = 0; otherFace < mlink.faces - 1;
						otherFace++) {
					int otherPos = otherFace * mlink.tiles + otherTile;
					if (pos != otherPos &&
							mlink.tileOfPosition[pos] - 1 == otherPos) {
						found = solveSpecialCases(pos, otherPos + 3);
					}
				}
			}
		}
	}

	void clearOut(int tile, int otherTile, int shift) {
		for (int face = 0; face < mlink.faces - 1; face++) {
			int pos = face * mlink.tiles + tile;
			for (int otherFace = 0; otherFace < mlink.faces - 1;
					otherFace++) {
				int otherPos = otherFace * mlink.tiles + otherTile;
				if (pos != otherPos &&
						mlink.tileOfPosition[pos] - 1 == otherPos) {
					solveSpecialCases(pos, pos + shift);
				}
			}
		}
	}

	void solveLeftMiddle() {
		solveMiddleFromEnd(0, 1);
		clearOut(1, 1, -1);
		solveMiddleFromEnd(0, 1);
		clearOut(3, 1, -3);
		solveMiddleFromEnd(0, 1);
		clearOut(2, 1, 1);
		clearOut(3, 1, -3);
		solveMiddleFromEnd(0, 1);
	}

	void solveLeft() {
		solveEnd(0);
		matchLeftRight(3, 0);
		solveEnds(3, 0);
		clearOut(2, 0, 1);
		matchLeftRight(3, 0);
		solveEnds(3, 0);
	}

	void solveRightMiddle() {
		solveMiddleFromEnd(3, 2);
		clearOut(2, 2, 1);
		solveMiddleFromEnd(3, 2);
	}

	void solveRight() {
		solveEnd(3);
	}

	/* This procedure coordinates the solution process. */
	void solveSomeTiles() {
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_RESET);
		if (!mlink.checkSolved()) {
			solveWhite();
			solveLeftMiddle();
			solveLeft();
			solveRightMiddle();
			solveRight();
		}
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_COMPUTED);
	}

	private class SeparateSubTask extends Thread {
		private boolean runFlag = false;
		SeparateSubTask() {
		}
		public void init() {
			start();
		}
		public void setFlag(boolean flag) {
			runFlag = flag;
		}
		public boolean getFlag() {
			return runFlag;
		}

		@Override
		public void run() {
			while (true) {
				try {
					sleep(100);
				} catch (InterruptedException e) {
					System.err.println("Interrupted");
				}
				if (runFlag) {
					solveSomeTiles();
					runFlag = false;
				}
			}
		}
	}

	private SeparateSubTask sp = null;
	public void setFlag(boolean flag) {
		if (sp != null)
			sp.setFlag(flag);
	}
	public boolean getFlag() {
		if (sp == null)
			return false;
		return sp.getFlag();
	}
	public void init() {
		if (sp == null) {
			sp = new SeparateSubTask();
			sp.init();
		}
	}
}
