package com.sillycycle.bagleyd.mlink;

/*
 * @(#)MlinkCanvas.java
 *
 * Copyright 1994 - 2022  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.URL;
import java.util.Random;
import javax.swing.JFrame;

/**
 * The <code>MlinkCanvas</code> class does all the graphics for
 * <code>MlinkFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class MlinkCanvas extends Canvas {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MAX_ITERATIONS = 64;
	static final int MAX_SLICES = 16;

	static final int TOP = 0;
	static final int RIGHT = 1;
	static final int BOTTOM = 2;
	static final int LEFT = 3;
	static final int COORD = 4;
	static final int THICKNESS = 8;
	static final int NORMAL = 1;
	static final int DOUBLE = 2;
	static final int INSTANT = 3;
	static final String DATAFILE = "mlink.dat";
	static final String SYMBOL = ":";
	static final String PICTUREFILE = MlinkInterface.IMAGE;
	static final int PICTUREX = 512;
	static final int PICTUREY = 512;

	Color       foreground, background;
	Color       tileColor, borderColor;
	Color       currentForeground;
	Color[]     faceColor = new Color[MlinkInterface.MAX_FACES];
	int[]       tileOfPosition = null;
	int[]       startPosition = null;
	int	 currentPosition = -1, currentSpace = -1;
	int	 currentDirection, lastDirection;
	int	 currentTile = -1, currentShift = -1;
	int	 nextPosition = -1, nextShift = -1;
	int	 columnPosition;
	int	 spacePosition, startSpace;
	boolean     started, cheat, motion;
	boolean     practice, orient, mono, reverse, middle;
	boolean     number = MlinkInterface.DEFAULT_NUMBER;
	int	 delay, numSlices, motionDiff, shiftDiff;
	int	 base;
	int	 lastX;
	int	 tiles, faces, tileFaces;
	Point       coreSize;
	Point       offset;
	Point       faceSize, tileSize, puzzleSize;
	Point       delta, puzzleOffset;

	JFrame      frame = null;
	int	 randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     movePaint = false, scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allTilesPaint = false;
	boolean     selectPaint = false, motionPaint = false;
	boolean     releasePaint = false;
	boolean     mouseDown = false;
	Random      generator;
	transient MlinkStack  undo = null, redo = null;
	transient MlinkSolve  solve = null;
	String      stringSave = null;
	String[]    token = null;
	String      faceNames = null;
	transient Image  bufferedFrameImage = null;
	transient Image[]  bufferedStartImage = new Image[2];
	transient Graphics  bufferedFrameGraphics = null;
	transient Graphics[]  bufferedStartGraphics = new Graphics[2];

	@Override
	protected String paramString() {
		return super.paramString() +
		",orient=" + orient +
		",middle=" + middle +
		",practice=" + practice +
		",tiles=" + tiles +
		",faces=" + faces +
		",base=" + base +
		",mono=" + mono +
		",reverse=" + reverse +
		",names=" + faceNames +
		",color0=" + faceColor[0] +
		",color1=" + faceColor[1] +
		",color2=" + faceColor[2] +
		",color3=" + faceColor[3] +
		",color4=" + faceColor[4] +
		",color5=" + faceColor[5] +
		",color6=" + faceColor[6] +
		",color7=" + faceColor[7] +
		",tileColor=" + tileColor +
		",borderColor=" + borderColor +
		",foreground=" + foreground +
		",background=" + background +
		",delay=" + delay;
	}

	void checkTiles() {
		if (tiles < MlinkInterface.MIN_TILES) {
			System.out.println(tiles +
			" number of mlink in a row incorrect, must be at least " +
			MlinkInterface.MIN_TILES +
			", defaulting to " +
			MlinkInterface.DEFAULT_TILES);
			tiles = MlinkInterface.DEFAULT_TILES;
		}
		if (faces < MlinkInterface.MIN_FACES ||
				faces > MlinkInterface.MAX_FACES) {
			System.out.println(faces +
			" number of mlink in a row incorrect, must be at between " +
			MlinkInterface.MIN_FACES + " and " +
			MlinkInterface.MAX_FACES + "inclusive, defaulting to " +
			MlinkInterface.DEFAULT_FACES);
			tiles = MlinkInterface.DEFAULT_FACES;
		}
		if (base < MlinkInterface.MIN_BASE ||
				base > MlinkInterface.MAX_BASE) {
			System.out.println(base +
			" base incorrect, must be between " +
			MlinkInterface.MIN_BASE +
			" and " +
			MlinkInterface.MAX_BASE +
			" inclusive, defaulting to " +
			MlinkInterface.DEFAULT_BASE);
			base = MlinkInterface.DEFAULT_BASE;
		}
		if (delay < 0) {
			System.out.println(delay +
			" delay cannot be negative, taking absolute value");
			delay = -delay;
		}
	}

	void setColumnPosition(int x) {
		columnPosition = (x - delta.x / 2 - puzzleOffset.x) / offset.x;
		if (columnPosition < 0 || columnPosition >= tiles)
			columnPosition = -1;
	}

	boolean positionTile(int x) {
		setColumnPosition(x);
		return (columnPosition >= 0 && columnPosition < tiles);
	}

	int positionToTile(int x, int y) {
		int temp = faces * offset.y + delta.y - 1;
		int i = (x - delta.x / 2 - puzzleOffset.x) /
			offset.x;
		int j = ((y - delta.y / 2 - puzzleOffset.y + temp) %
			temp) / offset.y;

		if (i >= 0 && j >= 0 && i < tiles && j < faces)
			return (i + tiles * j);
		return -1;
	}

	int toColumn(int pos) {
		return (pos % tiles);
	}

	int toRow(int pos) {
		return (pos / tiles);
	}

	int cartesianX(int pos) {
		return toColumn(pos) * offset.x + delta.x + puzzleOffset.x;
	}

	int cartesianY(int pos) {
		return toRow(pos) * (offset.y + 1) + delta.y + puzzleOffset.y;
	}

	int tileNFromSpace(int n, int direction) {
		int pos = spacePosition;

		if (direction == LEFT || direction == RIGHT)
			return (pos + ((direction == RIGHT) ? -n : n));
		/* direction == TOP || direction == BOTTOM */
		return (pos + (tiles * ((direction == BOTTOM) ? -n : n)));
	}

	boolean withinFrame(int x, int y, int dx, int dy) {
		return (x < dx + tileSize.x / 2 &&
			x > dx - tileSize.x / 2 &&
			tileSize.y * (x - dx) < tileSize.x * (y - dy) &&
			tileSize.y * (dx - x) < tileSize.x * (y - dy) &&
			tileSize.y * (x - dx + 2 * tileSize.x) >
				tileSize.x * (y - dy) &&
			tileSize.y * (dx - x + 2 * tileSize.x) >
				tileSize.x * (y - dy));
	}

	int solvedPosition(int tile) {
		return (tile + 1) % tileFaces;
	}

	boolean checkStarted() {
		return started;
	}

	boolean checkSolved() {
		int i, j, firstTile, lastTile, midTile;

		if (tiles < 2)
			return true;
		if (orient) {
			firstTile = tileOfPosition[0];
			if (firstTile != 0 && (firstTile - 1) % tiles == 0) {
				for (i = 1; i < tileFaces; i++) {
					midTile = tileOfPosition[i];
					if ((midTile != 0) && (midTile - firstTile +
							tileFaces) % tileFaces != i)
						return false;
				}
			} else
				return false;
		} else {
			for (i = 0; i < faces; i++) {
				firstTile = tileOfPosition[i * tiles];
				if (firstTile == 0)
					firstTile =
						tileOfPosition[i * tiles + 1];
				lastTile = tileOfPosition[(i + 1) * tiles - 1];
				if (lastTile == 0)
					lastTile = tileOfPosition[(i + 1) *
						tiles - 2];
				if (firstTile % tiles != 1 ||
						(lastTile % tiles != 0 &&
						lastTile != tileFaces - 1) ||
						(lastTile - 1) / tiles !=
						(firstTile - 1) / tiles)
					return false;
				for (j = 1; j < tiles - 1; j++) {
					midTile = tileOfPosition[i * tiles + j];
					if ((midTile - 1) / tiles !=
							(firstTile - 1) / tiles ||
							midTile == 0)
						return false;
				}
			}
		}
		started = false;
		return true;
	}

	void drawLink(Graphics g, int pos, int offsetX, int offsetY) {
		int dx, dy, sizex, sizey, tile = solvedPosition(pos);
		int thick;

		sizex = offset.x * 3 / 2;
		sizey = offset.y * 3 / 4;
		thick = java.lang.Math.min(sizey, sizex) / 3;
		thick = java.lang.Math.min(thick, THICKNESS);
		if (thick < 1)
			return;
		for (int i = 0; i < thick; i++) {
			sizex = offset.x * 3 / 2 - i;
			sizey = offset.y * 3 / 4 - i;
			dx = cartesianX(pos) - sizex / 2 + offsetX + 2;
			dy = cartesianY(pos) + tileSize.y / 2 - sizey / 2 -
				1 + offsetY;
			if (tile % tiles == 0 || tile == tileFaces - 1)
				g.drawArc(dx, dy, sizex, sizey, 89, -178);
			else if (tile % tiles == 1)
				g.drawArc(dx + tileSize.x - 1, dy,
					sizex, sizey, 91, 178);
			else {
				g.drawArc(dx, dy, sizex, sizey, 89, -32);
				g.drawArc(dx, dy, sizex, sizey, -89, 128);
				dx += tileSize.x - 1;
				g.drawArc(dx, dy, sizex, sizey, 91, 128);
				g.drawArc(dx, dy, sizex, sizey, -91, -33);
			}
		}
	}

	/* If only you could set line thickness in Java */
	/* void drawLink(int pos, int offsetX, int offsetY) {
		int dx, dy, sizex, sizey, tile = tileOfPosition[pos];
		int thick;

		sizex = offset.x * 3 / 2 - 3;
		sizey = offset.y * 3 / 4 - 3;
		dx = cartesianX(pos) - sizex / 2 + offsetX;
		dy = cartesianY(pos) + tileSize.y / 2 - sizey / 2 -
			1 + offsetY;
		thick = java.lang.Math.min(sizey, sizex) / 3;
		thick = java.lang.Math.min(thick, THICKNESS);
		if (thick > 1) {
			setLineAttributes(thick, LineSolid, CapNotLast,
				JoinRound);
			if (tile % tiles == 0 || tile == tileFaces - 1)
				g.drawArc(dx, dy, sizex, sizey, 89, -179);
			else if (tile % tiles == 1)
				g.drawArc(dx + tileSize.x - 1, dy,
					sizex, sizey, 90, 180);
			else {
				g.drawArc(dx, dy, sizex, sizey, 89, -124);
				g.drawArc(dx, dy, sizex, sizey, -90, 30);
				dx += tileSize.x - 1;
				g.drawArc(dx, dy, sizex, sizey, 90, 30);
				g.drawArc(dx, dy, sizex, sizey, -90, -125);
			}
			setLineAttributes(1, LineSolid, CapNotLast, JoinRound);
		}
	} */

	void copyImage(Graphics g, Image i,
			boolean scale,
			int dX, int dY, int sX, int sY,
			int width, int height) {
		if (scale) {
			g.drawImage(i, dX, dY, dX + width, dY + height,
				sX * PICTUREX / coreSize.x,
				sY * PICTUREY / coreSize.y,
				(sX + width) * PICTUREX / coreSize.x,
				(sY + height) * PICTUREY / coreSize.y, this);
		} else {
			g.drawImage(i, dX, dY, dX + width, dY + height,
				sX, sY, sX + width, sY + height, this);
		}
	}

	static void drawShadow(Graphics g, int startX, int startY,
			int sizeX, int sizeY) {
		g.drawLine(startX, startY, startX + sizeX, startY);
		g.drawLine(startX, startY, startX, startY + sizeY);
	}

	void drawTile(Graphics g, int pos,
			boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		int dx, dy, sx, sy;
		int tile = (tileOfPosition[pos] + tileFaces - 1) %
			tileFaces;

		dx = cartesianX(pos) + offsetX + pressedOffset;
		dy = cartesianY(pos) + offsetY + pressedOffset;
		sx = cartesianX(tile);
		sy = cartesianY(tile);
		if (pressedOffset == 0) {
			g.setColor(background);
			g.drawLine(dx + tileSize.x, dy,
				dx + tileSize.x, dy + tileSize.y);
			g.drawLine(dx, dy + tileSize.y,
				dx + tileSize.x, dy + tileSize.y);
		}
		if (blank) {
			g.setColor((erase) ? background : tileColor);
			g.fillRect(dx, dy, tileSize.x, tileSize.y);
			return;
		}
		copyImage(g, bufferedStartImage[pressedOffset], !number,
			dx, dy, sx, sy, tileSize.x, tileSize.y);
	}

	void drawTile(int pos, boolean blank, boolean erase,
			int pressedOffset, int offsetX, int offsetY) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawTile(g, pos, blank, erase,
				pressedOffset, offsetX, offsetY);
		} finally {
			g.dispose();
		}
	}

	void drawBufferedTile(Graphics g, int pos, int pressedOffset) {
		int dx, dy;
		int i, tile = solvedPosition(pos);
		int digitX = 1, digitY = 0, fontSize = 0;
		String buf = "";
		Color linkColor, fillColor, lineColor;

		if (g == null)
			return;
		dx = cartesianX(pos) + pressedOffset;
		dy = cartesianY(pos) + pressedOffset;
		if ((pressedOffset != 0) ^ reverse) {
			fillColor = borderColor;
			lineColor = tileColor;
		} else {
			fillColor = tileColor;
			lineColor = borderColor;
		}
		if (pressedOffset != 0) {
			g.setColor(fillColor.darker());
			drawShadow(g,
				dx - pressedOffset, dy - pressedOffset,
				tileSize.x - 1, tileSize.y - 1);
		}
		g.setColor(fillColor);
		g.fill3DRect(dx, dy, tileSize.x, tileSize.y,
			pressedOffset == 0);
		if (mono || orient) {
			g.setColor(lineColor);
			fontSize = java.lang.Math.min(tileSize.x,
				tileSize.y) / 2;
			g.setFont(new Font("arial", Font.PLAIN,
				fontSize));
			digitY = fontSize / 3;
		}
		if (orient) {
			i = tile;
			while (i >= 1) {
				i /= base;
				digitX += -fontSize / 3;
			}
			if (mono && (tile != 0 &&
					((tile != tileFaces - 1 &&
					tiles > 1) || tiles > 2)))
				buf = Integer.toString(tile, base).toUpperCase() +
					faceNames.charAt(((tile - 1) /
					tiles + 1) % faces);
			else
				buf = Integer.toString(tile, base).toUpperCase();
		} else if (mono && (tile != 0 &&
				((tile != tileFaces - 1 && tiles > 1) ||
				tiles > 2)))
			buf = Character.toString(faceNames.charAt(((tile - 1) /
				tiles + 1) % faces));
		if (mono && (tile != 0 &&
				((tile != tileFaces - 1 && tiles > 1) ||
				tiles > 2)))
			digitX += -fontSize / 3;
		if (mono && (tile != 0 &&
				((tile != tileFaces - 1 && tiles > 1) ||
				tiles > 2)) || orient)
			g.drawString(buf,
				dx + tileSize.x / 2 + digitX +
				tileSize.x / 25,
				dy + tileSize.y / 2 + digitY);
		if (mono) {
			linkColor = lineColor;
		} else {
			int face = ((tile - 1) / tiles + 1) % faces;
			linkColor = faceColor[face];
			if (((faceColor[face] == borderColor) &&
					pressedOffset != 0) ||
					((faceColor[face] == tileColor) &&
					pressedOffset == 0))
				linkColor = borderColor;
		}
		if (tile != 0 && ((tile != tileFaces - 1 && tiles > 1) ||
				tiles > 2)) {
			g.setColor(linkColor);
			drawLink(g, pos, pressedOffset - 1, pressedOffset + 1);
		}
		g.setColor(lineColor);
		g.drawRect(dx, dy, tileSize.x - 1, tileSize.y - 1);
	}

	void drawAllTiles() {
		for (int k = 0; k < tileFaces; k++) {
			drawTile(k, (tileOfPosition[k] <= 0),
				(tileOfPosition[k] <= 0), 0, 0, 0);
		}
	}

	void drawAllBufferedTiles() {
		if (number) {
			int k, l;

			for (k = 0; k < tileFaces; k++) {
				for (l = 0; l < 2; l++) {
					drawBufferedTile(
						bufferedStartGraphics[l],
						k, l);
				}
			}
		} else {
			URL url = MlinkFrame.class.getResource("/" + PICTUREFILE);

			if (url == null)
				bufferedStartImage[1] = null;
			else
				bufferedStartImage[1] = Toolkit.getDefaultToolkit().getImage(url);
			bufferedStartImage[0] = bufferedStartImage[1];
		}
	}

	void drawBuffer(Graphics g, Image i) {
		g.drawImage(i, 0, 0, coreSize.x, coreSize.y, this);
	}

	void drawBuffer(Image i) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawBuffer(g, i);
		} finally {
			g.dispose();
		}
	}

	void animateSlide(int numOfTiles, int dir,
			int fast, boolean logMoves) {
		Graphics g = getGraphics();
		int inc, aTile;
		int fillBeginPos, fillBeginI, gapI, moveI;
		int dx, dy, pos, posNext;
		int ix, iy;
		int numTiles = numOfTiles;

		if (g == null)
			return;
		try {
			fillBeginPos = tileNFromSpace(numTiles, dir);
			fillBeginI = cartesianX(fillBeginPos);
			//dy = cartesianY(fillBeginPos);
			gapI = tileSize.x * fast / numSlices;
			moveI = tileSize.x + delta.x;
			if (gapI == 0)
			  gapI++;
			if (numTiles < 0)
			  numTiles = -numTiles;
			for (inc = 0; inc < moveI + gapI; inc += gapI) {
			  if (inc > moveI)
			    inc = moveI;
			  for (aTile = 0; aTile < numTiles; aTile++) {
			    posNext = tileNFromSpace(aTile + 1, dir);
			    if (logMoves && inc == 0) {
			      //set direction for undo
			      ((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_MOVED);
			      undo.setMove(dir, posNext, 0);
			      redo.flushMoves();
			    }
			    /* Calculate deltas */
			    dx = cartesianX(posNext);
			    dy = cartesianY(posNext);
			    ix = ((dir == RIGHT) ? inc : -inc);
			    iy = 0;
			    drawTile(posNext, false, false, 0, ix, iy);
			    /* Erase old slivers */
			    ix += dx;
			    iy += dy;
			    g.setColor(background);
			    if (aTile < numTiles - 1) {
			      if (dir == RIGHT) {
				g.fillRect(ix - gapI, iy,
				  gapI, tileSize.y);
			      } else if (dir == LEFT) {
				g.fillRect(ix + tileSize.x, iy,
				  gapI, tileSize.y);
			      }
			    } else {
			      if (dir == RIGHT) {
				g.fillRect(fillBeginI, iy,
				  ix - fillBeginI, tileSize.y);
			      } else if (dir == LEFT) {
				g.fillRect(ix + tileSize.x, iy,
				  fillBeginI - ix, tileSize.y);
			      }
			    }
			  }
			  try {
			    Thread.sleep(delay / fast);
			  } catch (InterruptedException e) {
				  // ignore
			  }
			}
			pos = spacePosition;
			for (aTile = 0; aTile < numTiles; aTile++) {
			  posNext = tileNFromSpace(aTile + 1, dir);
			  tileOfPosition[pos] = tileOfPosition[posNext];
			  pos = posNext;
			}
			spacePosition = fillBeginPos;
			tileOfPosition[spacePosition] = 0;
		} finally {
			g.dispose();
		}
	}

	void drawAllTilesForColumns(Graphics g, int start, int finish,
			int dx, int dy) {
		int i, j, k;

		for (i = start; i <= finish; i++) {
			for (j = 0; j < faces; j++) {
				k = j * tiles + i;
				drawTile(g, k,
					(tileOfPosition[k] <= 0),
					(tileOfPosition[k] <= 0),
					0, -dx, -dy);
			}
		}
	}

	void animateRotate(int start, int finish, int direction, int fast) {
		Image columnImage;
		Graphics columnGraphics;
		Graphics g = getGraphics();
		int size = finish - start + 1;
		int moveI, gapJ, moveJ, inc, i, animateOffset = 0;

		if (g == null)
			return;
		try {
			moveI = tileSize.x + delta.x;
			gapJ = tileSize.y * fast / numSlices;
			moveJ = tileSize.y + delta.y + 1;
			if (gapJ == 0)
				gapJ++;
			columnImage = createImage(size * moveI,
				puzzleSize.y);
			columnGraphics = columnImage.getGraphics();
			copyImage(columnGraphics, bufferedFrameImage,
				false, 0, 0,
				start * moveI + puzzleOffset.x + 1,
				puzzleOffset.y,
				size * moveI, puzzleSize.y);
			drawAllTilesForColumns(columnGraphics, start, finish,
				(start * moveI + puzzleOffset.x + 1),
				puzzleOffset.y);
			for (inc = 0; inc < moveJ + gapJ; inc += gapJ) {
				if (inc > moveJ)
					inc = moveJ;
				if (direction == TOP) {
					i = inc;
				} else {
					i = puzzleSize.y - inc + 1;
					animateOffset = 3;
				}
				copyImage(g, columnImage,
					false,
					start * moveI + puzzleOffset.x + 1,
					puzzleOffset.y + animateOffset,
					0, i,
					size * moveI,
					puzzleSize.y - i - 1);
				copyImage(g, columnImage,
					false,
					start * moveI + puzzleOffset.x + 1,
					puzzleSize.y - i + puzzleOffset.y -
					delta.y - 1 + animateOffset,
					0, 0,
					size * moveI,
					i - delta.y - 1);
				try {
					Thread.sleep(delay / fast);
				} catch (InterruptedException e) {
					// ignore
				}
			}
		} finally {
			g.dispose();
		}
	}

	void drawFrame(Graphics g, boolean show) {
		int offsetX = puzzleOffset.x;
		int offsetY = puzzleOffset.y;

		if (g == null)
			return;
		if (show) {
			int sumX, sumY, k;

			g.setColor(borderColor);
			g.fillRect(0, 0, coreSize.x, offsetY);
			g.fillRect(0, coreSize.y - offsetY,
				coreSize.x, offsetY);
			g.fillRect(0, 0, offsetX, coreSize.y);
			g.fillRect(coreSize.x - offsetX, 0,
				offsetX, coreSize.y);
			g.setColor(currentForeground);
			sumX = tiles * offset.x + 1;
			sumY = offset.y + 1;
			for (k = 0; k < faces; k++) {
				g.fillRect(offsetX, offsetY, 1, sumY);
				g.fillRect(offsetX, offsetY, sumX, 1);
				g.fillRect(sumX + offsetX, offsetY,
					1, sumY + 1);
				offsetY += sumY;
				g.fillRect(offsetX, offsetY, sumX + 1, 1);
			}
		} else {
			/*
			g.setColor(Color.white);
			g.fillRect(0, 0, coreSize.x, offsetY);
			g.fillRect(0, coreSize.y - offsetY - 1,
				coreSize.x, offsetY + 1);
			g.fillRect(0, 0, offsetX, coreSize.y);
			g.fillRect(coreSize.x - offsetX - 1, 0,
				offsetX + 1, coreSize.y);
			*/
			g.setColor(background);
			g.fillRect(offsetX + 1, offsetY + 1,
				coreSize.x - 2 - 2 * offsetX,
				coreSize.y - 2 - 2 * offsetY);
		}
	}

	void drawFrame(boolean show) {
		Graphics g = getGraphics();

		if (g == null)
			return;
		try {
			drawFrame(g, show);
		} finally {
			g.dispose();
		}
	}

	void moveNoTiles() {
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_IGNORE);
	}

	void moveTiles(int from, int fast) {
		int l = from - spacePosition;

		if (fast != INSTANT && delay > 0) {
			if (l < 0) {
				animateSlide(-l, RIGHT, fast, false);
			} else {
				animateSlide(l, LEFT, fast, false);
			}
		} else {
			int pos, posNext, aTile, dir;

			dir = (l < 0) ? RIGHT : LEFT;
			if (l < 0)
				l = -l;
			pos = spacePosition;
			for (aTile = 0; aTile < l; aTile++) {
				posNext = tileNFromSpace(aTile + 1, dir);
				tileOfPosition[pos] = tileOfPosition[posNext];
				drawTile(pos, false, false, 0, 0, 0);
				pos = posNext;
			}
			spacePosition = from;
			tileOfPosition[from] = 0;
			drawTile(spacePosition, true, true, 0, 0, 0);
		}
		if (((MlinkFrame) frame).getToggleSound())
			((MlinkFrame) frame).playBumpAudio();
	}

	void rotateTiles(int start, int finish, int direction, int fast) {
		int newR, i, k, tempTile, currTile, pos, newPos;
		int r = 0, space = -1;
		boolean animate = (fast != INSTANT && delay > 0);

		if (animate)
			animateRotate(start, finish, direction, fast);
		for (i = start; i <= finish; i++) {
			currTile = tileOfPosition[r * tiles + i];
			for (k = 1; k <= faces; k++) {
				newR = (direction == TOP) ?
					(r + faces - 1) % faces :
					(r + 1) % faces;
				pos = r * tiles + i;
				newPos = newR * tiles + i;
				tempTile = tileOfPosition[newPos];
				tileOfPosition[newPos] = currTile;
				currTile = tempTile;
				if (spacePosition == pos) {
					if (!animate)
						drawTile(newPos,
							true, true, 0, 0, 0);
					space = newPos;
				} else if (!animate)
					drawTile(newPos,
						false, false, 0, 0, 0);
				r = newR;
			}
			if (space >= 0)
				spacePosition = space;
		}
		/* Errors at top when rotating down and bottom rotating up */
		if (animate) {
			Graphics g = getGraphics();

			if (g == null)
				return;
			try {
				//drawFrame(g, true);
				drawAllTilesForColumns(g, start, finish, 0, 0);
			} finally {
				g.dispose();
			}
		}
		if (((MlinkFrame) frame).getToggleSound())
			((MlinkFrame) frame).playMoveAudio();
	}

	boolean moveTilesDir(int direction, int tile, boolean all, int fast) {
		int start, finish;

		start = tile;
		finish = tile;
		if (all) {
			start = 0;
			finish = tiles - 1;
		} else if (middle && tile > 0 && tile < tiles - 1) {
			start = 1;
			finish = tiles - 2;
		}
		switch (direction) {
		case TOP:
			rotateTiles(start, finish, TOP, fast);
			return true;
		case RIGHT:
			if (tile >= 0 && toRow(spacePosition) ==
					toRow(tile)) {
				moveTiles(tile, fast);
				return true;
			}
			break;
		case BOTTOM:
			rotateTiles(start, finish, BOTTOM, fast);
			return true;
		case LEFT:
			if (toRow(spacePosition) == toRow(tile)) {
				moveTiles(tile, fast);
				return true;
			}
			break;
		default:
			System.out.println("moveTilesDir: direction " +
				direction);
		}
		return false;
	}

	boolean checkMoveTilesDir(int direction) {
		switch (direction) {
		case TOP:
			return true;
		case RIGHT:
			return (toColumn(spacePosition) > 0);
		case LEFT:
			return (toColumn(spacePosition) < tiles - 1);
		case BOTTOM:
			return true;
		default:
			System.out.println("checkMoveTilesDir: direction " +
				direction);
		}
		return false;
	}

	boolean exchangeTiles(int pos1, int pos2) {
		int tempTile;

		if (tileOfPosition[pos1] <= 0 ||
				tileOfPosition[pos2] <= 0)
			return false;
		tempTile = tileOfPosition[pos1];
		tileOfPosition[pos1] = tileOfPosition[pos2];
		tileOfPosition[pos2] = tempTile;
		return true;
	}

	boolean movePuzzle(int direction, int tile, boolean shift,
			boolean motionFlag, int fast) {
		if (shift && (direction == TOP || direction == BOTTOM)) {
			moveShiftCb(direction, fast);
			return true;
		}
		if (!checkMoveTilesDir(direction)) {
			return false;
		}
		started = true;
		if (!motionFlag) {
			if (direction == RIGHT) {
				int pos = spacePosition - tile;
				for (int n = pos % tiles - 1; n >= 0; n--) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_MOVED);
					undo.setMove(direction, tile + n, 0);
				}
			} else if (direction == LEFT) {
				int pos = tile - spacePosition;
				for (int n = pos % tiles - 1; n >= 0; n--) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_MOVED);
					undo.setMove(direction, tile - n, 0);
				}
			} else {
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_MOVED);
				undo.setMove(direction, tile, 0);
			}
			redo.flushMoves();
		}
		moveTilesDir(direction, tile, false, fast);
		return true;
	}

	void diffMove(int diff, int j, boolean shift) {
		int k;

		if (diff > faces / 2 || (diff == (faces + 1) / 2 &&
				j > currentPosition / tiles))
			for (k = 0; k < faces - diff; k++)
				movePuzzle(BOTTOM, currentPosition % tiles,
					shift, true, NORMAL);
		else
			for (k = 0; k < diff; k++)
				movePuzzle(TOP, currentPosition % tiles,
					shift, true, NORMAL);
	}

	void moveRecord(int direction, int tile, boolean shift) {
		if (shift && (direction == TOP || direction == BOTTOM)) {
			undo.setMove(direction, tile, 1);
		} else {
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_MOVED);
			undo.setMove(direction, tile, 0);
		}
		redo.flushMoves();
	}

	void diffRecord(int diff, int j, boolean shift) {
		int k;

		currentTile = toColumn(currentPosition);
		if (diff > faces / 2 || (diff == (faces + 1) / 2 &&
				j > currentPosition / tiles))
			for (k = 0; k < faces - diff; k++)
				moveRecord(BOTTOM,
					currentTile, shift);
		else
			for (k = 0; k < diff; k++)
				moveRecord(TOP,
					currentTile, shift);
	}

	void discreteSlide(int from, int dir) {
		moveTiles(from, NORMAL);
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_MOVED);
		undo.setMove(dir, from, 0);
		redo.flushMoves();
	}

	void slidePuzzle(int direction, int pos) {
		if (checkSolved() && !practice) {
			moveNoTiles();
			return;
		}
		if (!movePuzzle(direction, pos, false, false, NORMAL)) {
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_BLOCKED);
			return;
		}
		if (checkSolved()) {
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_SOLVED);
		}
	}

	void rotatePuzzle(int direction, int tile) {
		if (checkSolved() && !practice) {
			moveNoTiles();
			return;
		}
		movePuzzle(direction, tile, false, false, NORMAL);
		if (checkSolved()) {
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_SOLVED);
		}
	}

	void movePuzzleInput(int x, int direction,
			boolean shift, boolean control) {
		if (mouseDown || randomPaints != 0)
			return;
		if (stackPaint) {
			repaint();
			return;
		}
		if (control)
			controlPuzzle(direction);
		else if (shift && (direction == TOP || direction == BOTTOM)) {
			moveShiftCb(direction, NORMAL);
			undo.setMove(direction, 0, 1);
			redo.flushMoves();
		} else if (checkSolved() && !practice) {
			moveNoTiles();
		} else if (direction == LEFT || direction == RIGHT) {
			slidePuzzle(direction,
				spacePosition + direction - BOTTOM);
			if (checkSolved()) {
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_SOLVED);
			}
		} else {
			if (!positionTile(x))
				return;
			rotatePuzzle(direction, columnPosition);
			if (checkSolved()) {
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_SOLVED);
			}
		}
	}

	void controlPuzzle(int direction) {
		int reason = MlinkInterface.ACTION_IGNORE;
		switch (direction) {
		case TOP:
			if (faces >= MlinkInterface.MAX_FACES)
				return;
			reason = MlinkInterface.ACTION_INCY;
			break;
		case RIGHT:
			reason = MlinkInterface.ACTION_INCX;
			break;
		case BOTTOM:
			if (faces <= MlinkInterface.MIN_FACES)
				return;
			reason = MlinkInterface.ACTION_DECY;
			break;
		case LEFT:
			if (tiles <= MlinkInterface.MIN_TILES)
				return;
			reason = MlinkInterface.ACTION_DECX;
			break;
		default:
			System.out.println("ControlMlink: direction " +
				direction);
		}
		((MlinkFrame) frame).callback(reason);
	}

	void motionTiles() {
		if (nextShift == 0 && !practice && checkSolved()) {
			moveNoTiles();
			return;
		}
		if (!((toRow(nextPosition) == toRow(currentPosition)) &&
				(toRow(nextPosition) == toRow(spacePosition)))) {
			int j = nextPosition / tiles;
			int diff = (toRow(currentPosition) -
				toRow(nextPosition) + faces) % faces;

			diffMove(diff, j, nextShift != 0);
			if (nextShift == 0 && checkSolved()) {
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_SOLVED);
			}
			motion = true;
			if (nextShift == 0)
				motionDiff = (diff + motionDiff) % faces;
			else
				shiftDiff = (diff + shiftDiff) % faces;
		}
		currentPosition = nextPosition;
		if (nextShift != 0 || practice || !checkSolved()) {
			drawTile(currentPosition,
				(tileOfPosition[currentPosition] <= 0),
				false, 1, 0, 0);
		}
	}

	void selectTiles() {
		if (nextShift == 0 && !practice && checkSolved()) {
			moveNoTiles();
			return;
		}
		if ((toRow(nextPosition) == toRow(currentPosition)) &&
				(toRow(nextPosition) == toRow(spacePosition)) &&
				!motion) {
			if (currentPosition != spacePosition &&
					nextShift == 1 && checkSolved()) {
				moveNoTiles();
				return;
			}
			selectslideTiles(currentPosition - spacePosition);
			currentPosition = -1;
			started = true;
			if (checkSolved()) {
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_SOLVED);
			}
		} else {
			int j = nextPosition / tiles;
			int diff = (toRow(currentPosition) -
				toRow(nextPosition) + faces) % faces;

			currentTile = toColumn(currentPosition);
			diffMove(diff, j, nextShift != 0);
			if (nextShift == 0)
				motionDiff = (diff + motionDiff) % faces;
			else
				shiftDiff = (diff + shiftDiff) % faces;
			diffRecord(motionDiff, j, false);
			diffRecord(shiftDiff, j, true);
			if (nextShift == 0 && checkSolved()) {
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_SOLVED);
			}
			currentPosition = nextPosition;
			motionDiff = 0;
			shiftDiff = 0;
		}
	}

	void selectslideTiles(int pos) {
		if (pos < 0) {
			if (delay > 0) {
				animateSlide(-pos, RIGHT, NORMAL, true);
				if (((MlinkFrame) frame).getToggleSound())
					((MlinkFrame) frame).playBumpAudio();
			} else {
				for (int n = 1; n <= -pos % tiles; n++) {
					discreteSlide(spacePosition - 1,
						RIGHT);
				}
			}
		} else if (pos > 0) {
			if (delay > 0) {
				animateSlide(pos, LEFT, NORMAL, true);
				if (((MlinkFrame) frame).getToggleSound())
					((MlinkFrame) frame).playBumpAudio();
			} else {
				for (int n = 1; n <= pos % tiles; n++) {
					discreteSlide(spacePosition + 1,
						LEFT);
				}
			}
		}
	}

	void moveShiftCb(int direction, int fast) {
		moveTilesDir(direction, 0, true, fast);
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_CONTROL);
	}

	void resetTiles() {
		int i;

		solve.setFlag(false);
		tileFaces = tiles * faces;
		tileOfPosition = new int[tileFaces];
		startPosition = new int[tileFaces];
		spacePosition = tileFaces - 1;
		for (i = 1; i < tileFaces; i++)
			tileOfPosition[i - 1] = i;
		tileOfPosition[tileFaces - 1] = 0;
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		currentPosition = -1;
		currentDirection = -1;
		started = false;
		cheat = false;
		randomPaints = 0;
	}

	void practiceTiles() {
		((MlinkFrame) frame).callback(MlinkInterface.ACTION_PRACTICE);
	}

	void randomizeTiles() {
		if (practice)
			practiceTiles();
		/* First interchange tiles an even number of times */
		if (tiles > 1 && faces > 1 && tileFaces > 4) {
			int currentPos, randomPos;
			int count = 0;

			for (currentPos = 0; currentPos < tileFaces; currentPos++) {
				randomPos = currentPos;
				while (currentPos == randomPos)
					randomPos = generator.nextInt(tileFaces);
				if (exchangeTiles(currentPos, randomPos))
					count++;
			}
			if ((count & 1) == 1 && !exchangeTiles(0, 1) &&
					!exchangeTiles(tileFaces - 2,
					tileFaces - 1))
				System.out.println("randomizeTiles: should not get here");
			allTilesPaint = true;
			repaint();
		}
		/* Now move the spaces around randomly */
		if (tiles > 1 || faces > 1) {
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_RESET);
			lastDirection = -1;
			randomPaints = java.lang.Math.min(tileFaces,
				MAX_ITERATIONS) + generator.nextInt(2);
			repaint();
		} else if (checkSolved())
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_SOLVED);
	}

	void randomizingTiles() {
		currentDirection = generator.nextInt(COORD);
		if (((currentDirection + (COORD >> 1)) % COORD != lastDirection ||
				tiles == 1 || faces == 1) &&
				checkMoveTilesDir(currentDirection)) {
			movePuzzle(currentDirection, generator.nextInt(tiles),
				false, false, INSTANT);
			lastDirection = currentDirection;
			currentDirection = -1;
			randomPaints--;
			try {
				Thread.sleep(delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			repaint();
		} else {
			undo.flushMoves();
			redo.flushMoves();
			getStartPosition();
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_RANDOMIZE);
			started = false;
			currentPosition = -1;
			if (checkSolved())
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_SOLVED);
		}
	}

	public boolean getNumber() {
		return number;
	}

	public int getDelay() {
		return delay;
	}

	public void setDelay(int delay) {
		this.delay = delay;
		numSlices = ((delay < MAX_SLICES) ? delay + 1 : MAX_SLICES);
	}

	public boolean getCheat() {
		return cheat;
	}

	boolean setPuzzleOrient(boolean newValue) {
		if (orient != newValue) {
			orient = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleMiddle(boolean newValue) {
		if (middle != newValue) {
			middle = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzlePractice(boolean newValue) {
		if (practice != newValue) {
			practice = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleTiles(int newValue) {
		if (tiles != newValue && newValue >= MlinkInterface.MIN_TILES) {
			tiles = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	boolean setPuzzleFaces(int newValue) {
		if (faces != newValue && newValue >= MlinkInterface.MIN_FACES) {
			faces = newValue;
			resetTiles();
			resizePaint = true;
			repaint();
			return true;
		}
		return false;
	}

	public void speedUpPuzzle() {
		delay -= 10;
		if (delay < 0)
			delay = 0;
		((MlinkFrame) frame).callback(MlinkInterface.ACTION_SPEED);
	}

	public void slowDownPuzzle() {
		delay += 10;
		((MlinkFrame) frame).callback(MlinkInterface.ACTION_SPEED);
	}

	void showMessage(String msg) {
		((MlinkFrame) frame).showMessage(msg);
	}

	void resizePuzzle() {
		coreSize = new Point(getSize().width, getSize().height);
		delta = new Point(1, 1);
		offset = new Point(java.lang.Math.max((coreSize.x - 2) /
			tiles, 0),
			java.lang.Math.max(coreSize.y / faces - 2, 0));
		faceSize = new Point(offset.x * tiles + delta.x + 2,
			offset.y + delta.y + 2);
		puzzleSize = new Point(faceSize.x,
			(faceSize.y - 2) * faces + 2);
		puzzleOffset = new Point((coreSize.x - puzzleSize.x + 2) / 2,
			(coreSize.y - puzzleSize.y + 2) / 2);
		tileSize = new Point(java.lang.Math.max(offset.x - delta.x, 0),
			java.lang.Math.max(offset.y - delta.y, 0));
		bufferedFrameImage = createImage(coreSize.x, coreSize.y);
		bufferedFrameGraphics = bufferedFrameImage.getGraphics();
		if (number) {
			bufferedStartImage[0] =
				createImage(coreSize.x, coreSize.y);
			bufferedStartImage[1] =
				createImage(coreSize.x, coreSize.y);
		} else {
			bufferedStartImage[0] =
				createImage(PICTUREX, PICTUREY);
			bufferedStartImage[1] =
				createImage(PICTUREX, PICTUREY);
		}
		bufferedStartGraphics[0] = bufferedStartImage[0].getGraphics();
		bufferedStartGraphics[1] = bufferedStartImage[1].getGraphics();
	}

	public void initializePuzzle() {
		checkTiles();
		undo = new MlinkStack();
		redo = new MlinkStack();
		solve = new MlinkSolve(frame, this);
		resetTiles();
		generator = new Random(System.nanoTime());
	}

	void exposePuzzle() {
		drawFrame(bufferedFrameGraphics, false);
		drawFrame(bufferedFrameGraphics, true);
		drawBuffer(bufferedStartGraphics[0], bufferedFrameImage);
		drawBuffer(bufferedStartGraphics[1], bufferedFrameImage);
		drawAllBufferedTiles();
		drawBuffer(bufferedFrameImage);
		drawAllTiles();
		solve.init();
	}

	void selectPuzzle(boolean shift, int x, int y) {
		motion = false;
		motionDiff = 0;
		shiftDiff = 0;
		currentPosition = positionToTile(x, y);
		if (currentPosition >= 0) {
			if (!shift && !practice && checkSolved()) {
				moveNoTiles();
				currentPosition = -1;
				return;
			}
			selectPaint = true;
			repaint();
		} else
			currentPosition = -1;
	}

	void motionPuzzle(boolean shift, int x, int y) {
		if (currentPosition == -1)
			return;
		if (!shift && !practice && checkSolved()) {
			moveNoTiles();
			currentPosition = -1;
			return;
		}
		nextShift = (shift) ? 1 : 0;
		nextPosition = positionToTile(x, y);
		motionPaint = true;
		repaint();
	}

	void releasePuzzle(boolean shift, int x, int y) {
		if (currentPosition == -1)
			return;
		nextPosition = positionToTile(x, y);
		currentSpace = 0;
		nextShift = (shift) ? 1 : 0;
		releasePaint = true;
		repaint();
	}

	public void clearPuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		resetTiles();
		drawAllTiles();
		allTilesPaint = true;
		repaint();
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_RESET);
	}

	void randomizePuzzle() {
		if (!mouseDown)
			randomizeTiles();
	}

	void randomizePuzzleWithQuery() {
		if (!started)
			randomizePuzzle();
	}

	public void getPuzzle() {
		this.stringSave = ((MlinkFrame) frame).readTextArea();
		if (this.stringSave == null || this.stringSave.equals("")) {
			showMessage("Empty string to read");
			return;
		}
		getPuzzleString("buffer");
	}

	public void getPuzzleFile() {
		undo.flushMoves();
		redo.flushMoves();
		getStartPosition();
		try (BufferedReader br = new BufferedReader(
				new FileReader(DATAFILE))) {
			StringBuilder puzzleData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null)
				puzzleData.append(lineData).append(NL);
			this.stringSave = puzzleData.toString();
			if (this.stringSave == null || this.stringSave.equals("")) {
				showMessage("Empty file " + DATAFILE);
				return;
			}
			getPuzzleString(DATAFILE);
		} catch (IOException e) {
			//e.printStackTrace();
			showMessage("Cannot read from " + DATAFILE);
		}
	}

	public void getPuzzleString(String source) {
		try {
			int i, tilesValue, facesValue, moves;
			boolean orientValue, middleValue, practiceValue;
			String symbol = SYMBOL;
			String newline = NL;
			int line = 0;

			token = this.stringSave.split(newline);
			String lineData = token[line];
			if (lineData == null || lineData.length() == 0) {
				showMessage(source +
					" nothing in buffer");
				return;
			}
			tilesValue = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			if (tilesValue >= MlinkInterface.MIN_TILES) {
				for (i = this.tiles; i < tilesValue; i++) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_INCX);
				}
				for (i = this.tiles; i > tilesValue; i--) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_DECX);
				}
			} else {
				showMessage(source +
					" corrupted: tiles " + tilesValue +
					" should be at least " +
					MlinkInterface.MIN_TILES);
				return;
			}
			lineData = token[++line];
			facesValue = Integer.parseInt(lineData.substring(
					lineData.indexOf(symbol) + 2));
			if (facesValue >= MlinkInterface.MIN_FACES &&
					facesValue <= MlinkInterface.MAX_FACES) {
				for (i = this.faces; i < facesValue; i++) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_INCY);
				}
				for (i = this.faces; i > facesValue; i--) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_DECY);
				}
			} else {
				showMessage(source +
					" corrupted: faces " + facesValue +
					" should be at between " +
					MlinkInterface.MIN_FACES +
					" and " + MlinkInterface.MAX_FACES +
					" inclusive");
				return;
			}
			lineData = token[++line];
			middleValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.middle != middleValue)
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_MIDDLE);
			lineData = token[++line];
			orientValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.orient != orientValue)
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_ORIENTIZE);
			lineData = token[++line];
			practiceValue = (Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2)) != 0);
			if (this.practice != practiceValue)
				((MlinkFrame) frame).callback(
					MlinkInterface.ACTION_PRACTICE);
			lineData = token[++line];
			moves = Integer.parseInt(lineData.substring(
				lineData.indexOf(symbol) + 2));
			line = scanStartPosition(++line);
			if (line < 0)
				return;
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_RESTORE);
			setStartPosition();
			allTilesPaint = true;
			repaint();
			if (!scanMoves(++line, moves))
				return;
			System.out.println(source + ": tiles " + tilesValue +
				", faces " + facesValue + ", middle " + middleValue +
				", orient " + orientValue +
				", practice " + practiceValue + ", moves " + moves);
			cheat = true;
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading");
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Bad input file");
		}
	}

	public void printPuzzle(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append("tiles").append(symbol).append(space)
			.append(tiles).append(newline);
		sb.append("faces").append(symbol).append(space)
			.append(faces).append(newline);
		sb.append("middle").append(symbol).append(space)
			.append((middle) ? 1 : 0).append(newline);
		sb.append("orient").append(symbol).append(space)
			.append((orient) ? 1 : 0).append(newline);
		sb.append("practice").append(symbol).append(space)
			.append((practice) ? 1 : 0).append(newline);
		sb.append("moves").append(symbol).append(space)
			.append(undo.numMoves()).append(newline);
		printStartPosition(sb);
		printMoves(sb);
	}

	public void writePuzzle() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		((MlinkFrame) frame).writeTextArea(sb.toString());
	}

	public void writePuzzleFile() {
		StringBuilder sb = new StringBuilder();

		printPuzzle(sb);
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				DATAFILE, false))) {
			pr.print(sb);
			System.out.println("Saved to " + DATAFILE);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot write to " + DATAFILE);
		}
	}

	void undoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!undo.madeMoves())
			return;
		undo.getMove();
		currentDirection = undo.readMoveDirection();
		currentTile = undo.readmoveTile();
		currentShift = undo.readmoveShift();
		redo.setMove(currentDirection, currentTile, currentShift);
		// move in opposite direction
		currentDirection = (currentDirection + COORD / 2) % COORD;
		if (currentShift == 1 && (currentDirection == TOP ||
				currentDirection == BOTTOM)) {
			stackPaint = true;
			repaint();
		} else if (checkMoveTilesDir(currentDirection)) {
			if (currentDirection == LEFT)
				currentTile = spacePosition + 1;
			else if (currentDirection == RIGHT)
				currentTile = spacePosition - 1;
			stackPaint = true;
			repaint();
			started = true;
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_UNDO);
		} else {
			currentDirection = currentShift = currentTile = -1;
		}
	}

	void redoPuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0 ||
				!redo.madeMoves())
			return;
		redo.getMove();
		currentDirection = redo.readMoveDirection();
		currentTile = redo.readmoveTile();
		currentShift = redo.readmoveShift();
		undo.setMove(currentDirection, currentTile, currentShift);
		if (currentShift == 1 && (currentDirection == TOP ||
				currentDirection == BOTTOM)) {
			stackPaint = true;
			repaint();
		} else if (checkMoveTilesDir(currentDirection)) {
			if (currentDirection == LEFT)
				currentTile = spacePosition + 1;
			else if (currentDirection == RIGHT)
				currentTile = spacePosition - 1;
			stackPaint = true;
			repaint();
			started = true;
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_REDO);
		} else {
			currentDirection = currentShift = currentTile = -1;
		}
	}

	void solvePuzzle() {
		if (mouseDown || stackPaint || randomPaints != 0)
			return;
		if (tiles == 4 && faces == 4) {
			cheat = true;
			solve.setFlag(true);
			try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				System.err.println("Interrupted");
			}
			//solve.solveSomeTiles();
		} else {
			showMessage("Auto-solver: sorry, only implemented for" +
				NL + "4 tiles and 4 faces.");
		}
	}

	public void numberPuzzle() {
		number = !number;
		if (orient) {
			resizePaint = true;
			repaint();
		} else
			((MlinkFrame) frame).callback(
				MlinkInterface.ACTION_ORIENTIZE);
	}

	void orientizePuzzle() {
		if (mouseDown || randomPaints != 0 || !number)
			return;
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_ORIENTIZE);
	}

	void middlePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		((MlinkFrame) frame).callback(
			MlinkInterface.ACTION_MIDDLE);
	}

	void practicePuzzle() {
		if (mouseDown || randomPaints != 0)
			return;
		practiceTiles();
	}

	void practicePuzzleWithQuery() {
		if (!started)
			practicePuzzle();
	}

	void printPositions() {
		for (int i = 0; i < tileFaces; i++) {
			int pos = startPosition[i];
			if (pos < 100)
				System.out.print(" ");
			if (pos < 10 && pos >= 0)
				System.out.print(" ");
			System.out.print(pos);
			if (i % tiles == tiles - 1) {
				System.out.println();
			} else {
				System.out.print(" ");
			}
		}
		System.out.println();
	}

	void getStartPosition() {
		startSpace = spacePosition;
		System.arraycopy(tileOfPosition, 0, startPosition, 0, tileFaces);
	}

	void setStartPosition() {
		spacePosition = startSpace;
		System.arraycopy(startPosition, 0, tileOfPosition, 0, tileFaces); // drawAll
	}

	int scanStartPosition(int location) {
		String lineData;
		int line = location;

		try {
			String space = "\\s+";
			int n = 0;

			++line; //read startingPosition:
			for (int j = 0; j < faces; j++) {
				line++;
				if (line >= token.length) {
					System.out.println("Short data in StartPosition");
					return -1;
				}
				lineData = token[line];
				String[] st = lineData.split(space);
				int offset = ("".equals(st[0])) ? 1 : 0;
				for (int i = 0; i < tiles; i++) {
					int num = Integer.parseInt(st[i + offset].trim());

					startPosition[n] = num;
					if (num == 0)
						startSpace = n;
					n++;
				}
			}
			//blank line read
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found when reading (StartPosition)");
			return -1;
		}
		return line;
	}

	boolean scanMoves(int location, int moves) {
		Graphics g = getGraphics();

		if (g == null)
			return false;
		try {
			int l;
			int line = location;
			String lineData;
			String symbol = SYMBOL;
			String space = "\\s+";

			++line; //read moves direction
			paint(g);
			for (l = 0; l < moves && ++line < token.length; l++) {
				lineData = token[line];
				String[] st = lineData.substring(
					lineData.indexOf(symbol) + 2).split(space);
				currentDirection = Integer.parseInt(st[0].trim());
				currentTile = Integer.parseInt(st[1].trim());
				currentShift = Integer.parseInt(st[2].trim());
				if (checkMoveTilesDir(currentDirection)) {
					scanPaint = true;
					undo.setMove(currentDirection,
						currentTile, currentShift);
					redo.flushMoves();
					if (currentShift == 0) {
						((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_MOVED);
					}
					paint(g);
				} else {
					currentDirection = -1;
					System.out.println(l +
						": move in direction " +
						currentDirection +
						", cannot be made.");
					return false;
				}

			}
			if (l < moves) {
				showMessage("Corrupt input in scanMoves, only " + l + " moves");
				return false;
			}
		} catch (NumberFormatException e) {
			//e.printStackTrace();
			showMessage("Corrupt input found in scanMoves");
			return false;
		} finally {
			g.dispose();
		}
		return true;
	}

	void printStartPosition(StringBuilder sb) {
		String symbol = SYMBOL;
		String space = " ";
		String newline = NL;

		sb.append(newline);
		sb.append("startingPosition").append(symbol).append(newline);
		for (int i = 0; i < tileFaces; i++) {
			int pos = startPosition[i];
			if (pos < 100)
				sb.append(space);
			if (pos < 10 && pos >= 0)
				sb.append(space);
			sb.append(space).append(pos);
			if (((i + 1) % tiles) == 0) {
				sb.append(newline);
			}
		}
		sb.append(newline);
	}

	void printMoves(StringBuilder sb) {
		int direction, tile, shift;
		int nMoves = undo.numMoves();
		String symbol = SYMBOL;
		String space = "\t";
		String newline = NL;

		sb.append("moves").append(space)
			.append("dir").append(space)
			.append("tile").append(space)
			.append("shift").append(newline);
		while (undo.madeMoves()) {
			undo.getMove();
			direction = undo.readMoveDirection();
			shift = undo.readmoveShift();
			tile = undo.readmoveTile();
			redo.setMove(direction, tile, shift);
		}
		// apply all moves back to the stack since
		// we emptied it above!!
		for (int i = 0; i < nMoves; i++) {
			redo.getMove();
			direction = redo.readMoveDirection();
			shift = redo.readmoveShift();
			tile = redo.readmoveTile();
			sb.append(i + 1).append(symbol).append(space)
				.append(direction).append(space)
				.append(tile).append(space)
				.append(shift).append(newline);
			undo.setMove(direction, tile, shift);
		}
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allTilesPaint &&
				!selectPaint && !motionPaint &&
				!releasePaint && !movePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			drawFrame(bufferedFrameGraphics, true);
			framePaint = false;
		}
		if (allTilesPaint) {
			drawAllTiles();
			allTilesPaint = false;
		}
		if (selectPaint) {
			if (currentPosition != -1) {
				drawTile(currentPosition,
					(tileOfPosition[currentPosition] <= 0),
					false, 1, 0, 0);
			}
			selectPaint = false;
		}
		if (motionPaint) {
			if (currentPosition != -1) {
				int j = nextPosition / tiles;
				int k = currentPosition;

				drawTile(k, true, true, 1, 0, 0);
				drawTile(k, tileOfPosition[k] <= 0,
					tileOfPosition[k] <= 0, 0, 0, 0);
				diffRecord(motionDiff, j, false);
				diffRecord(shiftDiff, j, true);
				if (nextShift == 0 && checkSolved()) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_SOLVED);
				}
				if (tileOfPosition[currentPosition] > 0) {
					drawTile(currentPosition,
						false, false, 0, 0, 0);
				}
				motionDiff = 0;
				shiftDiff = 0;
				if (currentSpace != -1) {
					motionTiles();
				}
				currentPosition = nextPosition;
			}
			motionPaint = false;
		}
		if (releasePaint) {
			if (currentPosition != -1) {
				drawTile(currentPosition,
					tileOfPosition[currentPosition] <= 0,
					true, 1, 0, 0);
				if (tileOfPosition[currentPosition] > 0) {
					drawTile(currentPosition,
						false, false, 0, 0, 0);
				}
				if (currentSpace != -1) {
					selectTiles();
				}
				currentPosition = -1;
			}
			releasePaint = false;
		}
		if (movePaint || scanPaint || stackPaint) {
			int fast;

			if (stackPaint)
				fast = DOUBLE;
			else if (movePaint)
				fast = NORMAL;
			else
				fast = INSTANT;
			if (currentShift == 1 && (currentDirection == TOP ||
					currentDirection == BOTTOM)) {
				moveShiftCb(currentDirection, fast);
			} else {
				moveTilesDir(currentDirection,
					currentTile, false, fast);
				if (stackPaint && checkSolved()) {
					((MlinkFrame) frame).callback(
						MlinkInterface.ACTION_SOLVED);
				}
				started = true;
			}
			currentDirection = currentShift = currentTile = -1;
			movePaint = false;
			scanPaint = false;
			stackPaint = false;
		}
		if (randomPaints > 0) {
			randomizingTiles();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}
}
